#include <dmotor.h>
#include <unistd.h>
#include <dsensor.h>
#include <sys/lcd.h>
#include <dsound.h>
#include <lnp/lnp.h>
#include <lnp/lnp-logical.h>
#include "komm.h"

/*****************************************************************
** altalanos eljarasok & tipusok & makrok
**/
#define abs(a) ((a>0)?(a):(-a))

typedef unsigned char byte;

void beep(int i) {
  static note_t beepsound[] ={
	{ PITCH_G4, QUARTER },
	{ PITCH_END, 0 }
  };
  beepsound[0].pitch = i;
  dsound_play(beepsound);
}

#define ELSO_SZEM SENSOR_1

/*************************************************************************
** uzenetkuldes LNP-n 
**/

void UZENET( int mode, int msg)
{
  lnp_logical_range( mode );
  lnp_addressing_write( (byte*)& msg , sizeof(msg) , ELVESZETT_CIM, MENTO_PORT);
};


/**************************************************************************
**  erzekelok esemenyeinek kezelese
**/

int esemeny;

static wakeup_t touch_wakeup(wakeup_t data) 
{
/*	int i, e[2];
	e[0]=e[1]=0;
	for(i=0; i<5; ++i) { ++e[TOUCH(SENSOR_2)]; msleep(50); }
	return e[data]==5;*/
  return TOUCH(SENSOR_2)==data;
}

long elso_elozo, elso_villanas;
static wakeup_t villogas_wakeup()
{
  long elozo=elso_elozo,most=LIGHT(ELSO_SZEM);
  long voltvaltozas=abs(most-elozo)>5;

  elso_elozo=most;

  return voltvaltozas;
}

/**************
**  esemeny-sor: minden fenyvaltozas idobelyegzoje bekerul egy sorba,
**  a sorbol barmikor lekerdezhetjuk, hogy az elmult T idobn hany valtozas
**  tortent (check_blink). Akkor 'latjuk' a villogo fenyt, ha az elmult
**  1 mp-ben legalabb 4 fenyvaltozas (felvillanas/elsotetules) volt
**  (latom_a_fenyt). A fenyvaltozast egy kulon szal figyeli (pakolja a sorba az
**  idobelyegzoket...)
**/ 

#define CHECK_BLINK_TIME 1000
#define EVENT_ELVESZT 2
#define EVENT_BEAKASZT 1

#define MAX_CHANGE 100
static int elemszam = 0;
static time_t valtozasok[ MAX_CHANGE ];
static int v_eleje=0,v_vege=0;

int check_blink(time_t interval)
{
  time_t mar_regi;
  mar_regi = sys_time - interval;

  while( valtozasok[ v_eleje ] < mar_regi && elemszam>0)
  {
	v_eleje++;
	if(v_eleje >= MAX_CHANGE ) v_eleje=0;
	elemszam--;
  }

//  lcd_int( elemszam );
  return elemszam;  
}

int latom_a_fenyt()
{
  return check_blink(CHECK_BLINK_TIME) >= 4;
}


int villogas_figyelo_thread(int argc, char * argv[])
{
  elso_elozo=LIGHT( ELSO_SZEM );

  elso_villanas=0;

  for(;;)
  { 
    wait_event(&villogas_wakeup,0);

    elemszam++;
    valtozasok[ v_vege ] = sys_time;
    v_vege = (v_vege+1)%MAX_CHANGE;

  };

  return 0;
}

/**************************************************************************
** haladas es forgas 'makrok'
** regisztralja a mozgast (egy verembe pakol), ami alapjan a visszafele
** haladas megoldhato.
** a NE_REGISZTRAJ_TOBBET meghivando, ha megtalaltuk a cel-robotot.
** (kulonben tovabb gyujti a mozgast...)
**/

#define INV_IRANY(x) (((x)==fwd)?rev:fwd)
#define SGN_IRANY(x) (((x)==fwd)?1:(-1))
#define MAX_ALLAPOTOK 100

enum{ ALL, HALAD, FOROG};

int allapot;
int allapot_irany;
time_t allapot_kezdes;

typedef struct 
{
  int allapot;
  int allapot_irany;
  time_t ido;
} reg_allapot; 

reg_allapot allapot_verem[MAX_ALLAPOTOK];
int num_allapot=0;

void push_allapot( int all, int irany, time_t ido)
{
 beep(PITCH_G5);
 allapot_verem[ num_allapot ].ido = ido;
 allapot_verem[ num_allapot ].allapot = all;
 allapot_verem[ num_allapot ].allapot_irany = irany;
 num_allapot ++ ;
}

int regisztralj = 0;
void REGISZTRALJ_TOBBET(int i){ regisztralj=i;};

void REGISZTRAL_ALLAPOT()
{
	reg_allapot * reg_all;
	time_t ido = sys_time-allapot_kezdes;

	reg_all = & allapot_verem[ num_allapot - 1 ];

	if( reg_all->allapot == allapot)
	{
		if(reg_all->allapot_irany != allapot_irany)
		{
			if(reg_all->ido > ido)
			{
				reg_all->ido-=ido;
			}
			else
			{
				reg_all->ido = ido-reg_all->ido;
				reg_all->allapot_irany = allapot_irany;
			}
		}
		else
		{
		  reg_all->ido += ido;
		}
	}
	else
	{
		push_allapot( allapot, allapot_irany, ido);
	}

}


void halad(int irany)
{
	allapot = HALAD;
	allapot_irany = irany;
	allapot_kezdes = sys_time;

	motor_c_dir( irany );
	motor_c_speed( MAX_SPEED );
// blokkoljuk a forgast
	motor_a_speed( MAX_SPEED );
	motor_a_dir( brake );
}

void haladas_leall( )
{
	motor_c_speed( MAX_SPEED );
	motor_c_dir( brake );
}

void halad_ido(int irany,int ido)
{
	halad( irany );
	msleep( ido );
	haladas_leall();
}

#define jobb fwd
#define bal  rev

void forog(int irany)
{
	allapot = FOROG;
	allapot_irany = irany;
	allapot_kezdes = sys_time;

	motor_a_dir( irany );
	motor_a_speed( MAX_SPEED );
// blokkoljuk a haladast
	motor_c_speed( MAX_SPEED );
	motor_c_dir( brake );
}

void forgas_leall()
{
	motor_a_dir( MAX_SPEED );
	motor_a_dir( brake );
}

void forog_ido(int irany,int ido)
{
	forog( irany );
	msleep( ido );
	forgas_leall();
}

/***************************************************************************
** az elfogo-kar mikromotorjanak kodja
**/

#define MM_LE fwd
#define MM_IDO_LE 800
#define MM_IDO_FEL 1000

void kar_rogzit()
{

// a mikromotort elforgatjuk, hogy vizszintbe alljon az elfogo-kar
/*  motor_b_speed(MAX_SPEED/4);
  motor_b_dir( MM_LE );
  //wait_event( & touch_wakeup, 1 );
  msleep( MM_IDO_LE );
  motor_b_dir( off );
*/
// kezzel beallitjuk kezdetben
}

void kar_felenged()
{
	// a mikromotorral az elfogo-kart 'felengedjuk' kicsit:
	motor_b_speed(MAX_SPEED/4);
	motor_b_dir( INV_IRANY( MM_LE) );
	msleep( MM_IDO_FEL );
	motor_b_dir( off );
}

/****************
** egy wakeup fv, amit a main hasznal:
**/ 

wakeup_t elveszt_vagy_beakaszt_wakeup( wakeup_t w)
{
	if( touch_wakeup(0) ) {esemeny=EVENT_BEAKASZT; return EVENT_BEAKASZT;}
	if( !latom_a_fenyt() ) { esemeny=EVENT_ELVESZT; return EVENT_ELVESZT; }
	esemeny=0;
  return 0;
}

/***********************************************************************
************************************************************************
**
**	FOPROGRAM
**
**/

int main()
{
  int ujra;
  int pid_villogas_figyelo;
  time_t meglattam;
  int kar_felengedve;

  beep(PITCH_G3);
  msleep(500);

  kar_rogzit();

  ds_active( &ELSO_SZEM );

  lnp_logical_range(0);

  motor_a_speed(MAX_SPEED);
  motor_c_speed(MAX_SPEED);
  motor_c_speed( brake );
  motor_a_speed( brake );


  pid_villogas_figyelo = 
	  execi( & villogas_figyelo_thread, 0 , 0, 2, DEFAULT_STACK_SIZE);

//label raforgas:

  forog( jobb );

  do
  { 
	msleep(20);
  }while( ! latom_a_fenyt() );

  forgas_leall();			// forgas leall

  meglattam = sys_time;
  UZENET( HOSSZUTAV, NE_FOROGJ );



  push_allapot( ALL, 0,0);
  

  /* Elore haladas: */
  halad( fwd );

  /* Beakasztas: */

  ujra=1;
  
  while(ujra) {

    halad( fwd );

    /*esemeny = */wait_event( &elveszt_vagy_beakaszt_wakeup, 0 );

	if( esemeny == EVENT_ELVESZT)
	{
		time_t kezd;
        int messze;

  	    /* 
			visszahalad valamennyit
			kezdi elorol passziv erzekelovel
		*/

		haladas_leall();
    	REGISZTRAL_ALLAPOT();

		beep(PITCH_C4);

		// ha mar eleg kozel ert az elveszetthez, akkor visszatolat egy kicsit:
        if(sys_time - meglattam >= 2000) {
			messze=0;
			halad_ido( rev,2000 );
			REGISZTRAL_ALLAPOT();
		}
		else { 
			messze=1;
			meglattam = sys_time; // arra az esetre, ha az elejen sokszor elvesztene
		}

//       	ds_passive( &ELSO_SZEM );

		// kis ideig balra keressuk a fenyt, ha nincs meg, akkor jobbra
		// megkeressuk:
		kezd=sys_time;

		forog(bal);
	
		while(!latom_a_fenyt() && sys_time-kezd <= 3000) { msleep(50); }
		if(latom_a_fenyt()) {
			if(!messze) msleep(1000);
			else meglattam=sys_time;
			forgas_leall();
			REGISZTRAL_ALLAPOT();

		}
		else {
			forgas_leall();
			REGISZTRAL_ALLAPOT();
			forog(jobb);
			while(!latom_a_fenyt()) msleep(50);
			if(!messze) msleep(1000);
			forgas_leall();
			REGISZTRAL_ALLAPOT();

		}
	}

	else if( esemeny == EVENT_BEAKASZT)
	{

		beep( PITCH_G1 );
		msleep(500);
		haladas_leall();
		REGISZTRAL_ALLAPOT();
//		halad_ido( rev, 1000 );
/* meg kell varni, hogy az erzekelo-kar ne pattogjon... */
		msleep(250);	
	    if(touch_wakeup(1)) ujra=1; else ujra=0;
	}
  }


  // halad_ido( rev, 3000 );


  UZENET( HOSSZUTAV, NE_VILLOGJ );


  kar_felengedve = 0;

  // visszafele haladas, az elso muvelet forgas volt,azt nem kell vegrehajtani
  // itt: az utolso elem ( a veremben...)

  while(--num_allapot)
  {
	reg_allapot * regall = & allapot_verem[ num_allapot ];
	if(regall->allapot == HALAD)
	{
		halad_ido( INV_IRANY(regall->allapot_irany) , regall->ido);
	}
	else if (regall->allapot == FOROG)
	{
		// az elso forgasnal felengedjuk a kart, hogy tudjon forogni a robot
		if( ! kar_felengedve ) 
		{
			kar_felenged();
			kar_felengedve = 1;
		}

		forog_ido( INV_IRANY(regall->allapot_irany) , regall->ido);
	}
  }

  beep(PITCH_G4);

  // hogy ne brake legyen a vegen (mert az aramot fogyaszt es zug):
  motor_c_speed( off );
  motor_a_speed( off );

  return 0;
}
